@@ -5,7 +5,7 @@
  *	Lu-chuan Kung and Kang-pen Chen.
  *	All rights reserved.
  *
- * Copyright (c) 2004, 2005
+ * Copyright (c) 2004, 2005, 2006
  *	libchewing Core Team. See ChangeLog for details.
  *
  * See the file "COPYING" for information on usage and redistribution
@@ -14,17 +14,21 @@
 
 /**
  * @file chewingio.c
- * @brief Implement some routines for chewing
+ * @brief Implement basic I/O routines for Chewing manipulation.
  */
 
 #include <string.h>
 #include <ctype.h>
 
+#include "chewing-utf8-util.h"
 #include "chewingio.h"
 #include "global.h"
 #include "zuin.h"
 #include "chewingutil.h"
 #include "userphrase.h"
+#include "dict.h"
+#include "char.h"
+#include "hash.h"
 #include "private.h"
 
 #ifdef ENABLE_DEBUG
@@ -34,6 +38,8 @@
 FILE *fp_g;
 #endif
 
+extern int chewing_lifetime;
+
 void (*TerminateServices[ TerminateServicesNUM ])() = {
 	NULL
 };
@@ -49,10 +55,10 @@ static char *kb_type_str[] = {
 	"KB_ET26",
 	"KB_DVORAK",
 	"KB_DVORAK_HSU",
-	"KB_HANYU_PINYING"
+	"KB_HANYU_PINYIN"
 };
 
-int KBStr2Num( char str[] )
+CHEWING_API int chewing_KBStr2Num( char str[] )
 {
 	int i;
 
@@ -63,11 +69,6 @@ int KBStr2Num( char str[] )
 	return KB_DEFAULT;
 }
 
-void SetKBType( ZuinData *pZuin, int kbtype )
-{
-	pZuin->kbtype = kbtype;
-}
-
 #ifdef ENABLE_DEBUG     
 static void TerminateDebug()
 {
@@ -79,19 +80,53 @@ static void TerminateDebug()
 #endif
 
 int addTerminateService( void (*callback)() )
-{       
+{
 	if ( callback ) {
+		int i;
+		for ( i = 0; i < countTerminateService; ++i ) {
+			/* Avoid redundant function pointer */
+			if ( TerminateServices[ i ] == callback )
+				return 1;
+		}
 		TerminateServices[ countTerminateService++ ] = callback;
 		return 0;
 	}
 	return 1;
 }
 
-int InitChewing( void *iccf, ChewingConf *cf )
+CHEWING_API ChewingContext *chewing_new()
 {
-	ChewingData *pgdata = (ChewingData *) iccf;
+	ChewingContext *ctx;
+	
+	ChewingData *internal_data = ALC( ChewingData, 1 );
+	ChewingOutput *internal_output = ALC( ChewingOutput, 1 );
+	ctx = ALC( ChewingContext, 1 );
+	ctx->data = internal_data;
+	ctx->output = internal_output;
+	ctx->cand_no = 0;
+
+	/* handle configuration */
+	chewing_Reset( ctx );
+
+	return ctx;
+}
+
+CHEWING_API int chewing_Init(
+		const char *dataPath,
+		const char *hashPath )
+{
+	/* initialize Tree, Char, and Dict */
+	/* FIXME: check the validation of dataPath */
+	ReadTree( dataPath );
+	InitChar( dataPath );
+	InitDict( dataPath );
+
+	/* initial Hash */
+	/* FIXME: check the validation of hashPath */
+	ReadHash( hashPath );
 
 #ifdef ENABLE_DEBUG
+{
         char *dbg_path;
 	int failsafe = 1;
 	dbg_path = getenv( "CHEWING_DEBUG" );
@@ -109,14 +144,21 @@ int InitChewing( void *iccf, ChewingConf *cf )
 				"--> Output to stderr\n" );
 		}
 	}
+	/* register debug service */
 	if ( fp_g )
 		addTerminateService( TerminateDebug );
+}
 #endif
+	return 0;
+}
+
+CHEWING_API int chewing_Reset( ChewingContext *ctx )
+{
+	ChewingData *pgdata = ctx->data;
 
 	/* zuinData */
 	memset( &( pgdata->zuinData ), 0, sizeof( ZuinData ) );
 
-	SetKBType( &( pgdata->zuinData ), cf->kb_type );
 
 	/* choiceInfo */
 	memset( &( pgdata->choiceInfo ), 0, sizeof( ChoiceInfo ) );
@@ -128,16 +170,22 @@ int InitChewing( void *iccf, ChewingConf *cf )
 	memset( pgdata->bUserArrCnnct, 0, sizeof( int ) * ( MAX_PHONE_SEQ_LEN + 1 ) );
 	memset( pgdata->bUserArrBrkpt, 0, sizeof( int ) * ( MAX_PHONE_SEQ_LEN + 1 ) );
 	pgdata->bChiSym = CHINESE_MODE;
+	pgdata->bFullShape = HALFSHAPE_MODE;
 	pgdata->bSelect = 0;
 	pgdata->nSelect = 0;
 	pgdata->PointStart = -1;
 	pgdata->PointEnd = 0;
-	/* XXX: make options for callee to decide if use atexit or not. */
-	atexit( TerminateChewing );
+	pgdata->phrOut.nNumCut = 0;
 	return 0;
 }
 
-void TerminateChewing()
+CHEWING_API int chewing_set_KBType( ChewingContext *ctx, int kbtype )
+{
+	ctx->data->zuinData.kbtype = kbtype;
+	return 0;
+}
+
+CHEWING_API void chewing_Terminate()
 {
 	int i;
 
@@ -162,9 +210,20 @@ void TerminateChewing()
 	return;
 }
 
-int SetConfig( void *iccf, ConfigData *pcd )
+CHEWING_API void chewing_free( ChewingContext *ctx )
 {
-	ChewingData *pgdata = (ChewingData *) iccf;
+	if ( ctx->data )
+		free( ctx->data);
+	if ( ctx->output )
+		free( ctx->output);
+	if ( ctx )
+		free( ctx );
+	return;
+}
+
+CHEWING_API int chewing_Configure( ChewingContext *ctx, ChewingConfigData *pcd )
+{
+	ChewingData *pgdata = ctx->data;
 
 	pgdata->config.selectAreaLen = pcd->selectAreaLen;
 	pgdata->config.maxChiSymbolLen = pcd->maxChiSymbolLen;
@@ -174,18 +233,22 @@ int SetConfig( void *iccf, ConfigData *pcd )
 		sizeof( pcd->selKey[ 0 ] ) * MAX_SELKEY );
 	pgdata->config.bAddPhraseForward = pcd->bAddPhraseForward;
 	pgdata->config.bSpaceAsSelection = pcd->bSpaceAsSelection;
+	pgdata->config.bEscCleanAllBuf = pcd->bEscCleanAllBuf;
 	
 	/* Failback to default value */
 	if ( (pgdata->config.bAddPhraseForward != 0) && (pgdata->config.bAddPhraseForward != 1) )
 		pgdata->config.bAddPhraseForward = 0;
 	if ( (pgdata->config.bSpaceAsSelection != 0) && (pgdata->config.bSpaceAsSelection != 1) )
 		pgdata->config.bSpaceAsSelection = 1;
+	if ( (pgdata->config.bEscCleanAllBuf != 0) && (pgdata->config.bEscCleanAllBuf != 1) )
+		pgdata->config.bEscCleanAllBuf = 0;
+
 	return 0;
 }
 
-void SetChiEngMode( void *iccf, int mode )
+CHEWING_API void chewing_set_ChiEngMode( ChewingContext *ctx, int mode )
 {
-	ChewingData *pgdata = (ChewingData *) iccf;
+	ChewingData *pgdata = ctx->data;
 
 	if ( pgdata->bFirstKey == 0 ) {
 		pgdata->bChiSym = mode;
@@ -194,9 +257,19 @@ void SetChiEngMode( void *iccf, int mode )
 	}
 }
 
-int GetChiEngMode( void *iccf ) 
+CHEWING_API int chewing_get_ChiEngMode( ChewingContext *ctx ) 
 {
-	return ( (ChewingData *) iccf )->bChiSym;
+	return ctx->data->bChiSym;
+}
+
+CHEWING_API void chewing_set_ShapeMode( ChewingContext *ctx, int mode )
+{
+	ctx->data->bFullShape = (mode == FULLSHAPE_MODE ? 1 : 0);
+}
+
+CHEWING_API int chewing_get_ShapeMode( ChewingContext *ctx ) 
+{
+	return ctx->data->bFullShape;
 }
 
 static void CheckAndResetRange( ChewingData *pgdata )
@@ -213,13 +286,22 @@ static int DoSelect( ChewingData *pgdata, int num )
 		num += pgdata->choiceInfo.pageNo * pgdata->choiceInfo.nChoicePerPage;
 		/* Note: if num is larger than the total, there will be big troubles. */
 		if ( num < pgdata->choiceInfo.nTotalChoice ) {
-			if ( pgdata->choiceInfo.isSymbol == 1 ) {
+			if ( pgdata->choiceInfo.isSymbol ) {
 				SymbolChoice( pgdata, num );
 			} else { 
 				/* change the select interval & selectStr & nSelect */
 				AddSelect( pgdata, num );
 				/* second, call choice module */
 				ChoiceSelect( pgdata, num );
+				/* automatically shift the cursor to next phrase */
+				if ( pgdata->bAutoShiftCur != 0 &&
+				     /* if cursor at end of string, do not shift the cursor. */
+				     pgdata->chiSymbolCursor < pgdata->chiSymbolBufLen ) {
+					int len = pgdata->availInfo.avail[
+						pgdata->availInfo.currentAvail ].len;
+					pgdata->chiSymbolCursor += len;
+					pgdata->cursor += len;
+				}
 			}
 			return 1;
 		}
@@ -227,37 +309,85 @@ static int DoSelect( ChewingData *pgdata, int num )
 	return 0;
 }
 
-int OnKeySpace( void *iccf, ChewingOutput *pgo )
+CHEWING_API int chewing_handle_Space( ChewingContext *ctx )
 {
-	ChewingData *pgdata = (ChewingData *) iccf;
+	ChewingData *pgdata = ctx->data;
+	ChewingOutput *pgo = ctx->output;
 	int keystrokeRtn = KEYSTROKE_ABSORB;
 	int toSelect = 0;
-	int rtn;
+	int rtn, key_buf_cursor;
+	int bQuickCommit = 0;
 
 	/* check if Old Chewing style */
 	if ( ! pgdata->config.bSpaceAsSelection ) {
-		return OnKeyDefault( pgdata, ' ', pgo );
+		return chewing_handle_Default( ctx, ' ' );
 	}
 
 	CheckAndResetRange( pgdata );
 
 	if ( pgdata->bSelect ) {
 		if ( pgdata->choiceInfo.pageNo < ( pgdata->choiceInfo.nPage - 1 ) ) {
-			return OnKeyRight( iccf, pgo );
+			return chewing_handle_Right( ctx );
 		}
 	}
 
 	if ( ! ChewingIsEntering( pgdata ) ) {
-		rtn = SymbolInput( ' ', pgdata );
-		pgo->commitStr[ 0 ].wch = pgdata->chiSymbolBuf[ 0 ].wch;
+		if ( pgdata->bFullShape ) {
+			rtn = FullShapeSymbolInput( ' ', pgdata );
+		}
+		else {
+			rtn = SymbolInput( ' ', pgdata );
+		}
+		pgo->commitStr[ 0 ] = pgdata->chiSymbolBuf[ 0 ];
 		pgo->nCommitStr = 1;
 		pgdata->chiSymbolBufLen = 0;
 		pgdata->chiSymbolCursor = 0;
 		keystrokeRtn = KEYSTROKE_COMMIT;
-	} else if ( pgdata->bChiSym != CHINESE_MODE ) {
-		rtn = SymbolInput( ' ', pgdata );
+	} 
+	else if ( pgdata->bChiSym != CHINESE_MODE ) {
+		/* see if buffer contains nothing */
+		if ( pgdata->chiSymbolBufLen == 0 ) {
+			bQuickCommit = 1;
+		}
+
+		if ( pgdata->bFullShape ) {
+			rtn = FullShapeSymbolInput( ' ', pgdata );
+		}
+		else {
+			rtn = SymbolInput( ' ', pgdata );
+		}
+
 		keystrokeRtn = KEYSTROKE_ABSORB;
-	} else {
+		if ( rtn == SYMBOL_KEY_ERROR ) {
+			keystrokeRtn = KEYSTROKE_IGNORE;
+			/*
+			 * If the key is not a printable symbol, 
+			 * then it's wrongto commit it.
+			 */
+			bQuickCommit = 0;
+		} 
+		else {
+			keystrokeRtn = KEYSTROKE_ABSORB;
+		}
+
+		if ( ! bQuickCommit ) {
+                       CallPhrasing( pgdata );
+                       if( ReleaseChiSymbolBuf( pgdata, pgo ) != 0 )
+                               keystrokeRtn = KEYSTROKE_COMMIT;
+               }
+               /* Quick commit */
+               else {
+                       DEBUG_OUT(
+                               "\t\tQuick commit buf[0]=%c\n", 
+                               pgdata->chiSymbolBuf[ 0 ].s[ 0 ] );
+                       pgo->commitStr[ 0 ] = pgdata->chiSymbolBuf[ 0 ]; 
+                       pgo->nCommitStr = 1;
+                       pgdata->chiSymbolBufLen = 0;
+                       pgdata->chiSymbolCursor = 0;
+                       keystrokeRtn = KEYSTROKE_COMMIT;
+               }
+	}
+	else {
 		rtn = ZuinPhoInput( &( pgdata->zuinData ), ' ' );
 		switch ( rtn ) {
 			case ZUIN_ABSORB:
@@ -272,21 +402,13 @@ int OnKeySpace( void *iccf, ChewingOutput *pgo )
 				break;
 			case ZUIN_KEY_ERROR:
 			case ZUIN_IGNORE:
-				if ( pgdata->chiSymbolCursor == pgdata->chiSymbolBufLen ) {
-					if (
-						ChewingIsChiAt(
-							pgdata->chiSymbolCursor - 1,
-							pgdata ) ) {
-						toSelect = 1;
-					}
-				} else {
-					if (
-						ChewingIsChiAt(
-							pgdata->chiSymbolCursor,
-							pgdata ) ) {
-						toSelect = 1;
-					}
-				}
+				key_buf_cursor = pgdata->chiSymbolCursor;
+				if ( pgdata->chiSymbolCursor == pgdata->chiSymbolBufLen )
+					key_buf_cursor--;
+
+				/* see if to select */
+				if ( ChewingIsChiAt( key_buf_cursor, pgdata ) )
+					toSelect = 1;
 
 				if ( toSelect ) {
 					if ( ! pgdata->bSelect )
@@ -294,6 +416,11 @@ int OnKeySpace( void *iccf, ChewingOutput *pgo )
 					else
 						ChoiceNextAvail( pgdata );
 				}
+				else if ( pgdata->symbolKeyBuf[ key_buf_cursor ] ) {
+					/* Open Symbol Choice List */
+					if( ! pgdata->choiceInfo.isSymbol )
+						OpenSymbolChoice( pgdata );
+				}
 				break;
 		}
 	}
@@ -302,9 +429,10 @@ int OnKeySpace( void *iccf, ChewingOutput *pgo )
 	return 0;
 }
 
-int OnKeyEsc( void *iccf, ChewingOutput *pgo )
+CHEWING_API int chewing_handle_Esc( ChewingContext *ctx )
 {
-	ChewingData *pgdata = (ChewingData *) iccf;
+	ChewingData *pgdata = ctx->data;
+	ChewingOutput *pgo = ctx->output;
 	int keystrokeRtn = KEYSTROKE_ABSORB;
 
 	CheckAndResetRange( pgdata );
@@ -322,15 +450,18 @@ int OnKeyEsc( void *iccf, ChewingOutput *pgo )
 		ChoiceEndChoice( pgdata );
 	} else if ( ZuinIsEntering( &( pgdata->zuinData ) ) ) {
 		ZuinRemoveAll( &( pgdata->zuinData ) );
+	} else if( pgdata->config.bEscCleanAllBuf ) {
+		CleanAllBuf( pgdata );
 	}
 
 	MakeOutputWithRtn( pgo, pgdata, keystrokeRtn );
 	return 0;
 }
 
-int OnKeyEnter( void *iccf, ChewingOutput *pgo )
+CHEWING_API int chewing_handle_Enter( ChewingContext *ctx )
 {
-	ChewingData *pgdata = (ChewingData *) iccf;
+	ChewingData *pgdata = ctx->data;
+	ChewingOutput *pgo = ctx->output;
 	int nCommitStr = pgdata->chiSymbolBufLen;
 	int keystrokeRtn = KEYSTROKE_ABSORB;
 
@@ -338,24 +469,24 @@ int OnKeyEnter( void *iccf, ChewingOutput *pgo )
 		keystrokeRtn = KEYSTROKE_IGNORE;
 	} else if ( pgdata->bSelect ) {
 		keystrokeRtn = KEYSTROKE_ABSORB | KEYSTROKE_BELL;
-	} else if ( pgdata->PointStart > -1) {
+	} else if ( pgdata->PointStart > -1 ) {
 		int buf = pgdata->cursor;
 		int key;
 		if ( pgdata->PointEnd > 0 ) {
-			if ( !pgdata->config.bAddPhraseForward ) {
+			if ( ! pgdata->config.bAddPhraseForward ) {
 				pgdata->cursor = pgdata->PointStart;
 				key = '0' + pgdata->PointEnd;
 			} else {
 				pgdata->cursor = pgdata->PointStart + pgdata->PointEnd;
 				key = '0' + pgdata->PointEnd;
 			}
-			OnKeyCtrlNum( (void *) pgdata, key, pgo );
+			chewing_handle_CtrlNum( ctx, key );
 			pgdata->cursor = buf;
 		} else if ( pgdata->PointEnd < 0 ) {
 			if ( pgdata->config.bAddPhraseForward )
 				pgdata->cursor = buf - pgdata->PointEnd;
 			key = '0' - pgdata->PointEnd;
-			OnKeyCtrlNum( (void *) pgdata, key, pgo );
+			chewing_handle_CtrlNum( ctx, key );
 			pgdata->cursor = buf;
 		}
 		pgdata->PointStart = -1;
@@ -365,7 +496,6 @@ int OnKeyEnter( void *iccf, ChewingOutput *pgo )
 		WriteChiSymbolToBuf( pgo->commitStr, nCommitStr, pgdata );
 		AutoLearnPhrase( pgdata );
 		CleanAllBuf( pgdata );  
-		CallPhrasing( pgdata );
 		pgo->nCommitStr = nCommitStr;
 	}
 
@@ -373,9 +503,10 @@ int OnKeyEnter( void *iccf, ChewingOutput *pgo )
 	return 0;
 }
 
-int OnKeyDel( void *iccf, ChewingOutput *pgo )
+CHEWING_API int chewing_handle_Del( ChewingContext *ctx )
 {
-	ChewingData *pgdata = (ChewingData *) iccf;
+	ChewingData *pgdata = ctx->data;
+	ChewingOutput *pgo = ctx->output;
 	int keystrokeRtn = KEYSTROKE_ABSORB;
 
 	CheckAndResetRange( pgdata );
@@ -400,9 +531,10 @@ int OnKeyDel( void *iccf, ChewingOutput *pgo )
 	return 0;
 }
 
-int OnKeyBackspace( void *iccf, ChewingOutput *pgo )
+CHEWING_API int chewing_handle_Backspace( ChewingContext *ctx )
 {
-	ChewingData *pgdata = (ChewingData *) iccf;
+	ChewingData *pgdata = ctx->data;
+	ChewingOutput *pgo = ctx->output;
 	int keystrokeRtn = KEYSTROKE_ABSORB;
 
 	CheckAndResetRange( pgdata );
@@ -429,9 +561,10 @@ int OnKeyBackspace( void *iccf, ChewingOutput *pgo )
 	return 0;
 }
 
-int OnKeyUp( void *iccf, ChewingOutput *pgo )
+CHEWING_API int chewing_handle_Up( ChewingContext *ctx )
 {
-	ChewingData *pgdata = (ChewingData *) iccf;
+	ChewingData *pgdata = ctx->data;
+	ChewingOutput *pgo = ctx->output;
 	int keystrokeRtn = KEYSTROKE_ABSORB;
 
 	CheckAndResetRange( pgdata );
@@ -447,11 +580,13 @@ int OnKeyUp( void *iccf, ChewingOutput *pgo )
 	return 0;
 }
 
-int OnKeyDown( void *iccf, ChewingOutput *pgo )
+CHEWING_API int chewing_handle_Down( ChewingContext *ctx )
 {
-	ChewingData *pgdata = (ChewingData *) iccf;
+	ChewingData *pgdata = ctx->data;
+	ChewingOutput *pgo = ctx->output;
 	int toSelect = 0;
 	int keystrokeRtn = KEYSTROKE_ABSORB;
+	int key_buf_cursor;
 
 	CheckAndResetRange( pgdata );
 
@@ -459,14 +594,13 @@ int OnKeyDown( void *iccf, ChewingOutput *pgo )
 		keystrokeRtn = KEYSTROKE_IGNORE;
 	}
 
+	key_buf_cursor = pgdata->chiSymbolCursor;
+	if ( pgdata->chiSymbolCursor == pgdata->chiSymbolBufLen )
+		key_buf_cursor--;
+
 	/* see if to select */
-	if ( pgdata->chiSymbolCursor == pgdata->chiSymbolBufLen ) {
-		if ( ChewingIsChiAt( pgdata->chiSymbolCursor - 1, pgdata ) )
+	if ( ChewingIsChiAt( key_buf_cursor, pgdata ) )
 			toSelect = 1;
-	} else {
-		if ( ChewingIsChiAt( pgdata->chiSymbolCursor, pgdata ) ) 
-			toSelect = 1;
-	}
 
 	if ( toSelect ) {
 		if( ! pgdata->bSelect ) {
@@ -474,6 +608,11 @@ int OnKeyDown( void *iccf, ChewingOutput *pgo )
 		} else {
 			ChoiceNextAvail( pgdata );
 		}
+	} 
+	else if ( pgdata->symbolKeyBuf[ key_buf_cursor ] ) {
+		/* Open Symbol Choice List */
+		if ( ! pgdata->choiceInfo.isSymbol )
+			OpenSymbolChoice( pgdata );
 	}
 
 	MakeOutputWithRtn( pgo, pgdata, keystrokeRtn );
@@ -481,20 +620,21 @@ int OnKeyDown( void *iccf, ChewingOutput *pgo )
 }
 
 /* Add phrase in Hanin Style */
-int OnKeyShiftLeft( void *iccf, ChewingOutput *pgo )
+CHEWING_API int chewing_handle_ShiftLeft( ChewingContext *ctx )
 {
-	ChewingData *pgdata = (ChewingData *)iccf ;
-	int keystrokeRtn = KEYSTROKE_ABSORB ;
+	ChewingData *pgdata = ctx->data;
+	ChewingOutput *pgo = ctx->output;
+	int keystrokeRtn = KEYSTROKE_ABSORB;
 
 	if ( ! ChewingIsEntering( pgdata ) ) {
-		keystrokeRtn = KEYSTROKE_IGNORE ;
+		keystrokeRtn = KEYSTROKE_IGNORE;
 	} 
-	if(!pgdata->bSelect ) {
+	if ( ! pgdata->bSelect ) {
 		/*  PointEnd locates (-9, +9) */
 		if ( 
 			! ZuinIsEntering( &( pgdata->zuinData ) ) && 
-			pgdata->chiSymbolCursor > 0
-			&& pgdata->PointEnd > -9 ) {
+			pgdata->chiSymbolCursor > 0 &&
+			pgdata->PointEnd > -9 ) {
 			pgdata->chiSymbolCursor--;
 			if ( pgdata->PointStart == -1 )
 				pgdata->PointStart = pgdata->cursor;
@@ -513,9 +653,10 @@ int OnKeyShiftLeft( void *iccf, ChewingOutput *pgo )
 	return 0;
 }
 
-int OnKeyLeft( void *iccf, ChewingOutput *pgo )
+CHEWING_API int chewing_handle_Left( ChewingContext *ctx )
 {
-	ChewingData *pgdata = (ChewingData *) iccf;
+	ChewingData *pgdata = ctx->data;
+	ChewingOutput *pgo = ctx->output;
 	int keystrokeRtn = KEYSTROKE_ABSORB;
 
 	if ( ! ChewingIsEntering( pgdata ) ) {
@@ -545,16 +686,17 @@ int OnKeyLeft( void *iccf, ChewingOutput *pgo )
 }
 
 /* Add phrase in Hanin Style */
-int OnKeyShiftRight(void *iccf, ChewingOutput *pgo)
+CHEWING_API int chewing_handle_ShiftRight( ChewingContext *ctx )
 {
-	ChewingData *pgdata = (ChewingData *)iccf ;
-	int keystrokeRtn = KEYSTROKE_ABSORB ;
+	ChewingData *pgdata = ctx->data;
+	ChewingOutput *pgo = ctx->output;
+	int keystrokeRtn = KEYSTROKE_ABSORB;
 
-	if( ! ChewingIsEntering(pgdata)) {
-		keystrokeRtn = KEYSTROKE_IGNORE ;
+	if ( ! ChewingIsEntering( pgdata ) ) {
+		keystrokeRtn = KEYSTROKE_IGNORE;
 	} 
 
-	if( ! pgdata->bSelect) {
+	if ( ! pgdata->bSelect ) {
 		/* PointEnd locates (-9, +9) */
 		if ( 
 			! ZuinIsEntering( &( pgdata->zuinData ) ) && 
@@ -578,9 +720,10 @@ int OnKeyShiftRight(void *iccf, ChewingOutput *pgo)
 	return 0;
 }
 
-int OnKeyRight( void *iccf, ChewingOutput *pgo )
+CHEWING_API int chewing_handle_Right( ChewingContext *ctx )
 {
-	ChewingData *pgdata = (ChewingData *) iccf;
+	ChewingData *pgdata = ctx->data;
+	ChewingOutput *pgo = ctx->output;
 	int keystrokeRtn = KEYSTROKE_ABSORB;
 
 	if ( ! ChewingIsEntering( pgdata ) ) {
@@ -610,9 +753,10 @@ int OnKeyRight( void *iccf, ChewingOutput *pgo )
 	return 0;
 }
 
-int OnKeyTab( void *iccf, ChewingOutput *pgo )
+CHEWING_API int chewing_handle_Tab( ChewingContext *ctx )
 {
-	ChewingData *pgdata = (ChewingData *) iccf;
+	ChewingData *pgdata = ctx->data;
+	ChewingOutput *pgo = ctx->output;
 	int keystrokeRtn = KEYSTROKE_ABSORB;
 
 	CheckAndResetRange( pgdata );
@@ -623,7 +767,10 @@ int OnKeyTab( void *iccf, ChewingOutput *pgo )
 
 
 	if ( ! pgdata->bSelect ) {
-		if ( ChewingIsChiAt( pgdata->chiSymbolCursor - 1, pgdata ) ) {
+		if ( pgdata->chiSymbolCursor == pgdata->chiSymbolBufLen ) {
+			pgdata->phrOut.nNumCut++;
+		}
+		else if ( ChewingIsChiAt( pgdata->chiSymbolCursor - 1, pgdata ) ) {
 			if ( IsPreferIntervalConnted( pgdata->cursor, pgdata) ) {
 				pgdata->bUserArrBrkpt[ pgdata->cursor ] = 1;
 				pgdata->bUserArrCnnct[ pgdata->cursor ] = 0;
@@ -639,9 +786,10 @@ int OnKeyTab( void *iccf, ChewingOutput *pgo )
 	return 0;
 }
 
-int OnKeyDblTab( void *iccf, ChewingOutput *pgo )
+CHEWING_API int chewing_handle_DblTab( ChewingContext *ctx )
 {
-	ChewingData *pgdata = (ChewingData *) iccf;
+	ChewingData *pgdata = ctx->data;
+	ChewingOutput *pgo = ctx->output;
 	int keystrokeRtn = KEYSTROKE_ABSORB;
 
 	CheckAndResetRange( pgdata );
@@ -661,9 +809,10 @@ int OnKeyDblTab( void *iccf, ChewingOutput *pgo )
 }
 
 
-int OnKeyCapslock( void *iccf, ChewingOutput *pgo )
+CHEWING_API int chewing_handle_Capslock( ChewingContext *ctx )
 {
-	ChewingData *pgdata = (ChewingData *) iccf;
+	ChewingData *pgdata = ctx->data;
+	ChewingOutput *pgo = ctx->output;
 
 	pgdata->bChiSym = 1 - pgdata->bChiSym;
 	pgdata->bCaseChange = ( pgdata->bChiSym == CHINESE_MODE ? 0 : 1 );
@@ -671,9 +820,10 @@ int OnKeyCapslock( void *iccf, ChewingOutput *pgo )
 	return 0;
 }
 
-int OnKeyHome( void *iccf, ChewingOutput *pgo )
+CHEWING_API int chewing_handle_Home( ChewingContext *ctx )
 {
-	ChewingData *pgdata = (ChewingData *) iccf;
+	ChewingData *pgdata = ctx->data;
+	ChewingOutput *pgo = ctx->output;
 	int keystrokeRtn = KEYSTROKE_ABSORB;
 
 	CheckAndResetRange( pgdata );
@@ -689,9 +839,10 @@ int OnKeyHome( void *iccf, ChewingOutput *pgo )
 	return 0;
 }
 
-int OnKeyEnd( void *iccf, ChewingOutput *pgo )
+CHEWING_API int chewing_handle_End( ChewingContext *ctx )
 {
-	ChewingData *pgdata = (ChewingData *) iccf;
+	ChewingData *pgdata = ctx->data;
+	ChewingOutput *pgo = ctx->output;
 	int keystrokeRtn = KEYSTROKE_ABSORB;
 
 	CheckAndResetRange( pgdata );
@@ -735,19 +886,27 @@ static int dvorak_convert( int key )
 	return key;
 }
 
-int OnKeyDefault( void *iccf, int key, ChewingOutput *pgo )
+CHEWING_API int chewing_handle_Default( ChewingContext *ctx, int key )
 {
-	ChewingData *pgdata = (ChewingData *) iccf;
+	ChewingData *pgdata = ctx->data;
+	ChewingOutput *pgo = ctx->output;
 	int rtn, num;
 	int keystrokeRtn = KEYSTROKE_ABSORB;
 	int bQuickCommit = 0;
 
+	/* Update lifetime */
+	chewing_lifetime++;
+
 	/* Skip the special key */
 	if ( key & 0xFF00 ) {
 		keystrokeRtn = KEYSTROKE_IGNORE;
-		goto End_OnKeyDefault;
+		goto End_KeyDefault;
 	}
 
+	/* We ignore non-printable input */
+	if ( ! isprint( key ) )
+		goto End_KeyDefault;
+
 	CheckAndResetRange( pgdata );
 
 	DEBUG_CHECKPOINT();
@@ -761,10 +920,56 @@ int OnKeyDefault( void *iccf, int key, ChewingOutput *pgo )
 	/* selecting */
 	if ( pgdata->bSelect ) {
 		if ( key == ' ' )
-			return OnKeyRight( iccf, pgo );
+			return chewing_handle_Right( ctx );
 		/* num starts from 0 */
 		num = CountSelKeyNum( key, pgdata );
-		DoSelect( pgdata, num );
+		if ( num >= 0 ) {
+			DoSelect( pgdata, num );
+			goto End_KeyDefault;
+		}
+		
+		/* Otherwise, use 'j' and 'k' for paging in selection mode */
+		DEBUG_OUT(
+			"\t\tchecking paging key, got '%c'\n",
+			key );
+		switch ( key ) {
+			case 'j':
+			case 'J':
+				if ( pgdata->chiSymbolCursor > 0 ) {
+					if ( ! ChewingIsEntering( pgdata ) ) {
+						keystrokeRtn = KEYSTROKE_IGNORE;
+					}
+					CheckAndResetRange( pgdata );
+					pgdata->chiSymbolCursor--;
+					if ( (pgdata->cursor > 0) && 
+					     ChewingIsChiAt(
+						     pgdata->chiSymbolCursor, 
+						     pgdata ) ) {
+						pgdata->cursor--;
+					}
+					ChoiceFirstAvail( pgdata );
+				}
+				goto End_Paging;
+			case 'k':
+			case 'K':
+				if ( pgdata->chiSymbolCursor < pgdata->chiSymbolBufLen ) {
+					if ( ! ChewingIsEntering( pgdata ) ) {
+						keystrokeRtn = KEYSTROKE_IGNORE;
+					}
+					CheckAndResetRange( pgdata );
+					if ( (pgdata->cursor < pgdata->nPhoneSeq) &&
+					     ChewingIsChiAt(
+						     pgdata->chiSymbolCursor, 
+						     pgdata ) ) {
+						pgdata->cursor++;
+					}
+					pgdata->chiSymbolCursor++;
+					ChoiceFirstAvail( pgdata );
+				}
+				goto End_Paging;
+			default:
+				break;
+		}
 	}
 	/* editing */
 	else {
@@ -806,7 +1011,18 @@ int OnKeyDefault( void *iccf, int key, ChewingOutput *pgo )
 						bQuickCommit = 1;
 					}
 
-					rtn = SymbolInput( key, pgdata );
+					if ( pgdata->bFullShape ) {
+						rtn = FullShapeSymbolInput( key, pgdata );
+					}
+					else {
+						rtn = SymbolInput( key, pgdata );
+					}
+					if ( pgdata->bFullShape ) {
+						rtn = FullShapeSymbolInput( key, pgdata );
+					}
+					else {
+						rtn = SymbolInput( key, pgdata );
+					}
 					if ( rtn == SYMBOL_KEY_ERROR ) {
 						keystrokeRtn = KEYSTROKE_IGNORE;
 						/*
@@ -827,7 +1043,13 @@ int OnKeyDefault( void *iccf, int key, ChewingOutput *pgo )
 			if ( pgdata->chiSymbolBufLen == 0 ) {
 				bQuickCommit = 1;
 			}
-			rtn = SymbolInput( key, pgdata );
+			if ( pgdata->bFullShape ) {
+				rtn = FullShapeSymbolInput( key, pgdata );
+			}
+			else {
+				rtn = SymbolInput( key, pgdata );
+			}
+
 			if ( rtn == SYMBOL_KEY_ERROR ) {
 				keystrokeRtn = KEYSTROKE_IGNORE;
 				bQuickCommit = 0;
@@ -843,7 +1065,7 @@ int OnKeyDefault( void *iccf, int key, ChewingOutput *pgo )
 			DEBUG_OUT(
 				"\t\tQuick commit buf[0]=%c\n", 
 				pgdata->chiSymbolBuf[ 0 ].s[ 0 ] );
-			pgo->commitStr[ 0 ].wch = pgdata->chiSymbolBuf[ 0 ].wch; 
+			pgo->commitStr[ 0 ] = pgdata->chiSymbolBuf[ 0 ]; 
 			pgo->nCommitStr = 1;
 			pgdata->chiSymbolBufLen = 0;
 			pgdata->chiSymbolCursor = 0;
@@ -851,20 +1073,31 @@ int OnKeyDefault( void *iccf, int key, ChewingOutput *pgo )
 		}
 	}
 
-End_OnKeyDefault:
+	if ( pgdata->phrOut.nNumCut > 0 ) {
+		int i;
+		for ( i = 0; i < pgdata->phrOut.nDispInterval; i++ ) {
+			pgdata->bUserArrBrkpt[ pgdata->phrOut.dispInterval[ i ].from ] = 1;
+			pgdata->bUserArrBrkpt[ pgdata->phrOut.dispInterval[ i ].to ] = 1;
+		}
+		pgdata->phrOut.nNumCut = 0;
+	}
+
+End_KeyDefault:
 	CallPhrasing( pgdata );
+End_Paging:
 	MakeOutputWithRtn( pgo, pgdata, keystrokeRtn );
 	return 0;
 }
 
-int OnKeyCtrlNum( void *iccf, int key, ChewingOutput *pgo )
+CHEWING_API int chewing_handle_CtrlNum( ChewingContext *ctx, int key )
 {
-	ChewingData *pgdata = (ChewingData *) iccf;
+	ChewingData *pgdata = ctx->data;
+	ChewingOutput *pgo = ctx->output;
 	int keystrokeRtn = KEYSTROKE_ABSORB;
 	int newPhraseLen;
 	int i;
 	uint16 addPhoneSeq[ MAX_PHONE_SEQ_LEN ];
-	char addWordSeq[ MAX_PHONE_SEQ_LEN * 2 + 1 ];
+	char addWordSeq[ MAX_PHONE_SEQ_LEN * MAX_UTF8_SIZE + 1 ];
 	int phraseState;
 
 	CheckAndResetRange( pgdata );
@@ -879,7 +1112,7 @@ int OnKeyCtrlNum( void *iccf, int key, ChewingOutput *pgo )
 			&( pgdata->availInfo ), 
 			pgdata->phoneSeq, 
 			pgdata->config.selectAreaLen ); 
-		SemiSymbolInput('1', pgdata);
+		SemiSymbolInput( '1', pgdata );
                 CallPhrasing( pgdata );
                 MakeOutputWithRtn( pgo, pgdata, keystrokeRtn );
                 return 0;
@@ -894,16 +1127,15 @@ int OnKeyCtrlNum( void *iccf, int key, ChewingOutput *pgo )
 				pgdata->cursor, 
 				pgdata->cursor + newPhraseLen - 1 ) ) {
 				/* Manually add phrase to the user phrase database. */
-				memcpy(
-					addPhoneSeq,
-					&pgdata->phoneSeq[ pgdata->cursor ],
-					sizeof( uint16 ) * newPhraseLen );
+				memcpy( addPhoneSeq,
+				        &pgdata->phoneSeq[ pgdata->cursor ],
+				        sizeof( uint16 ) * newPhraseLen );
 				addPhoneSeq[ newPhraseLen ] = 0;
-				memcpy(
-					addWordSeq,
-					&pgdata->phrOut.chiBuf[ pgdata->cursor * 2 ],
-					sizeof( char ) * 2 * newPhraseLen );
-				addWordSeq[ newPhraseLen * 2 ] = '\0';
+				ueStrNCpy( addWordSeq,
+				           ueStrSeek( (char *) &pgdata->phrOut.chiBuf,
+				                      pgdata->cursor ),
+				           newPhraseLen, 1);
+
 
 				phraseState = UserUpdatePhrase( addPhoneSeq, addWordSeq );
 				SetUpdatePhraseMsg( 
@@ -922,21 +1154,18 @@ int OnKeyCtrlNum( void *iccf, int key, ChewingOutput *pgo )
 		if ( 
 			newPhraseLen >= 1 && 
 			pgdata->cursor - newPhraseLen >= 0 ) {
-			if ( NoSymbolBetween( 
-				pgdata, 
-				pgdata->cursor, 
-				pgdata->cursor - newPhraseLen ) ) {
+			if ( NoSymbolBetween( pgdata, 
+			                      pgdata->cursor, 
+			                      pgdata->cursor - newPhraseLen ) ) {
 				/* Manually add phrase to the user phrase database. */
-				memcpy(
-					addPhoneSeq,
-					&pgdata->phoneSeq[ pgdata->cursor - newPhraseLen ],
-					sizeof( uint16 ) * newPhraseLen );
+				memcpy( addPhoneSeq,
+				        &pgdata->phoneSeq[ pgdata->cursor - newPhraseLen ],
+				        sizeof( uint16 ) * newPhraseLen );
 				addPhoneSeq[ newPhraseLen ] = 0;
-				memcpy(
-					addWordSeq,
-					&pgdata->phrOut.chiBuf[ ( pgdata->cursor - newPhraseLen ) * 2 ],
-					sizeof( char ) * 2 * newPhraseLen );
-				addWordSeq[ newPhraseLen * 2 ] = '\0';
+				ueStrNCpy( addWordSeq,
+				           ueStrSeek( (char *) &pgdata->phrOut.chiBuf,
+				           pgdata->cursor - newPhraseLen ),
+				           newPhraseLen, 1);
 
 				phraseState = UserUpdatePhrase( addPhoneSeq, addWordSeq );
 				SetUpdatePhraseMsg( 
@@ -957,9 +1186,10 @@ int OnKeyCtrlNum( void *iccf, int key, ChewingOutput *pgo )
 	return 0;
 }
 
-int OnKeyCtrlOption( void *iccf, int key, ChewingOutput *pgo )
+CHEWING_API int chewing_handle_CtrlOption( ChewingContext *ctx, int key )
 {
-	ChewingData *pgdata = (ChewingData *)iccf;
+	ChewingData *pgdata = ctx->data;
+	ChewingOutput *pgo = ctx->output;
 	int rtn;
 	int keystrokeRtn = KEYSTROKE_ABSORB;
 
@@ -972,9 +1202,10 @@ int OnKeyCtrlOption( void *iccf, int key, ChewingOutput *pgo )
 	return 0;
 }
 
-int OnKeyShiftSpace( void *iccf, ChewingOutput *pgo )
+CHEWING_API int chewing_handle_ShiftSpace( ChewingContext *ctx )
 {
-	ChewingData *pgdata = (ChewingData *) iccf;
+	ChewingData *pgdata = ctx->data;
+	ChewingOutput *pgo = ctx->output;
 	int rtn, key = ' ';
 	int keystrokeRtn = KEYSTROKE_ABSORB;
 
@@ -987,9 +1218,10 @@ int OnKeyShiftSpace( void *iccf, ChewingOutput *pgo )
 	return 0;
 }
 
-int OnKeyNumlock( void *iccf, int key, ChewingOutput *pgo )
+CHEWING_API int chewing_handle_Numlock( ChewingContext *ctx, int key )
 {
-	ChewingData *pgdata = (ChewingData *) iccf;
+	ChewingData *pgdata = ctx->data;
+	ChewingOutput *pgo = ctx->output;
 	int rtn, QuickCommit = 0;
 	int keystrokeRtn = KEYSTROKE_ABSORB;
 	
@@ -1001,24 +1233,32 @@ int OnKeyNumlock( void *iccf, int key, ChewingOutput *pgo )
 			QuickCommit = 1;
 		}
 		rtn = SymbolInput( key, pgdata );
-		/* copied from OnKey Default */
+		/* copied from chewing_handle_Default */
 		if ( rtn == SYMBOL_KEY_ERROR ) {
 			keystrokeRtn = KEYSTROKE_IGNORE ;
 		} else if ( QuickCommit ) {
-			pgo->commitStr[ 0 ].wch = pgdata->chiSymbolBuf[ 0 ].wch; 
+			pgo->commitStr[ 0 ] = pgdata->chiSymbolBuf[ 0 ]; 
 			pgo->nCommitStr = 1;
 			pgdata->chiSymbolBufLen = 0;
 			pgdata->chiSymbolCursor = 0;
 			keystrokeRtn = KEYSTROKE_COMMIT;
 		}
+		else	/* Not quick commit */
+		{
+			CallPhrasing( pgdata );
+			if( ReleaseChiSymbolBuf( pgdata, pgo ) != 0 )
+				keystrokeRtn = KEYSTROKE_COMMIT;
+		}
 	} else {
 		/* Otherwise, if we are selecting words, we use numeric keys
 		 * as selkey
 		 * and submit the words. 
 		 */
 		int num = -1;
-		if ( key > '0' && key < '8' )
+		if ( key > '0' && key < '9' )
 			num = key - '1';
+		else if ( key == '0' )
+			num = 9;
 		DoSelect( pgdata, num );
  	}
 	CallPhrasing( pgdata );
