@@ -5,7 +5,7 @@
  *	Lu-chuan Kung and Kang-pen Chen.
  *	All rights reserved.
  *
- * Copyright (c) 2004, 2005
+ * Copyright (c) 2004, 2005, 2006
  *	libchewing Core Team. See ChangeLog for details.
  *
  * See the file "COPYING" for information on usage and redistribution
@@ -16,52 +16,46 @@
 #include <string.h>
 #include <stdio.h>
 
+#include "chewing-utf8-util.h"
 #include "global.h"
 #include "chewingutil.h"
 #include "zuin.h"
 #include "userphrase.h"
 #include "private.h"
 
-extern const char *ph_pho[]; 
+extern const char *zhuin_tab[]; 
 static void MakePreferInterval( ChewingData *pgdata );
 static void ShiftInterval( ChewingOutput *pgo, ChewingData *pgdata );
 
 void SetUpdatePhraseMsg( ChewingData *pgdata, char *addWordSeq, int len, int state )
 {
-	char *insert = "[J", *modify = "w";
+	char *insert = "加入－", *modify = "已有－";
 	int begin = 3, i;
 
 	pgdata->showMsgLen = begin + len;
 	if ( state == USER_UPDATE_INSERT ) {
-		pgdata->showMsg[ 0 ].s[ 0 ] = insert[ 0 ];
-		pgdata->showMsg[ 0 ].s[ 1 ] = insert[ 1 ];
-		pgdata->showMsg[ 0 ].s[ 2 ] = 0;
-		pgdata->showMsg[ 1 ].s[ 0 ] = insert[ 2 ];
-		pgdata->showMsg[ 1 ].s[ 1 ] = insert[ 3 ];
-		pgdata->showMsg[ 1 ].s[ 2 ] = 0;
-		pgdata->showMsg[ 2 ].s[ 0 ] = insert[ 4 ];
-		pgdata->showMsg[ 2 ].s[ 1 ] = insert[ 5 ];
-		pgdata->showMsg[ 2 ].s[ 2 ] = 0 ;
+		ueStrNCpy( (char *) pgdata->showMsg[ 0 ].s, insert, 1, 1);
+		ueStrNCpy( (char *) pgdata->showMsg[ 1 ].s,
+		           ueStrSeek( insert, 1 ), 
+		           1, 1 );
+		ueStrNCpy( (char *) pgdata->showMsg[ 2 ].s,
+		           ueStrSeek( insert, 2 ), 
+		           1, 1 );
 	}
 	else {
-		pgdata->showMsg[ 0 ].s[ 0 ] = modify[ 0 ];
-		pgdata->showMsg[ 0 ].s[ 1 ] = modify[ 1 ];
-		pgdata->showMsg[ 0 ].s[ 2 ] = 0;
-		pgdata->showMsg[ 1 ].s[ 0 ] = modify[ 2 ];
-		pgdata->showMsg[ 1 ].s[ 1 ] = modify[ 3 ];
-		pgdata->showMsg[ 1 ].s[ 2 ] = 0;
-		pgdata->showMsg[ 2 ].s[ 0 ] = modify[ 4 ];
-		pgdata->showMsg[ 2 ].s[ 1 ] = modify[ 5 ];
-		pgdata->showMsg[ 2 ].s[ 2 ] = 0;
+		ueStrNCpy( (char *) pgdata->showMsg[ 0 ].s, modify, 1, 1);
+		ueStrNCpy( (char *) pgdata->showMsg[ 1 ].s,
+		           ueStrSeek( modify, 1 ),
+			   1, 1 );
+		ueStrNCpy( (char *) pgdata->showMsg[ 2 ].s,
+		           ueStrSeek( modify, 2 ),
+			   1, 1 );
 	}
 	for ( i = 0; i < len; i++ ) {
-		pgdata->showMsg[ begin + i ].s[ 0 ] = addWordSeq[ i * 2 ];
-		pgdata->showMsg[ begin + i ].s[ 1 ] = addWordSeq[ i * 2 + 1 ];
-		pgdata->showMsg[ begin + i ].s[ 2 ] = 0 ;
+		ueStrNCpy( (char *) pgdata->showMsg[ begin + i ].s,
+		           ueStrSeek(addWordSeq, i),
+			   1, 1);
 	}
-	pgdata->showMsg[ begin + i].s[ 0 ] = 0;
-	pgdata->showMsg[ begin + i].s[ 1 ] = 0;
-	pgdata->showMsg[ begin + i].s[ 2 ] = 0;
 }
 
 int NoSymbolBetween( ChewingData *pgdata, int begin, int end )
@@ -93,53 +87,60 @@ int ChewingIsEntering( ChewingData *pgdata )
 
 #define CEIL_DIV(a,b) ((a+b-1)/b)
 
-int HaninSymbolInput(ChoiceInfo *pci, AvailInfo *pai, const uint16 phoneSeq[],	int selectAreaLen) {
-    static char *chibuf[] = { "A","B","C","D","F","G",
-                              "H","I","J","?","K","L","U","X", "[","\\",
-		"]","^","_","`","m","n","o","p",
-		"q","r","s","t","i","j","k","l",
-		"a","b","c","d","e","f","g","h",
-		"u","v","w","x","y","z","{","|",
-		"","","","","","","","",
-		"","","","","","","","",
-		"","","","","","","","",
-		"","","","","","","","",
-		"O","P","Q","R","S","T","U","V",
-		"W","J","K","X","","","","",
-		"","","","","","","","",
-		"","","","?","z","s","{","u",
-		"q","t","|","r","}","w","x","",
-		"","","","","","","~","",
-		"","","b","c","d","e","f","g",
-		"h","i","j","k","l","m","n","o",
-		"p","v","","","","","","",
-		"","","","Q","","","","",
-		"","","","","","","","",
-		"D","E","F","G","H","I","J","K",
-		"L","M","N","O","P","Q","R","S",
-		"T","U","V","W","X","Y","Z","[",
-		"\\","]","^","_","`","a","b","c",
-                              "d","e","f","g","h","i","j","k",
-                              "l","m","n","o","p","q","r","s"};
-    int i, all = 216;
+int HaninSymbolInput(
+		ChoiceInfo *pci, 
+		AvailInfo *pai, 
+		const uint16 phoneSeq[],
+		int selectAreaLen)
+{
+	static char *chibuf[] = { 
+		"，","、","。","．","；","：",
+		"？","！","︰","?","…","‥","｜","—", "︴","﹏",
+		"（","）","︵","︶","《","》","︽","︾",
+		"〈","〉","︿","﹀","【","】","︻","︼",
+		"｛","｝","︷","︸","〔","〕","︹","︺",
+		"「","」","﹁","﹂","『","』","﹃","﹄",
+		"＃","＆","＊","※","§","〃","○","●",
+		"◎","㊣","△","▽","▲","▼","∴","∵",
+		"☆","★","◇","◆","□","■","♀","♂",
+		"→","←","↑","↓","↗","↖","↘","↙",
+		"㏕","㎜","㎝","㎞","㏎","㎡","㎎","㎏",
+		"㏄","℃","℉","°","±","×","÷","≒",
+		"≠","≦","≧","～","∠","⊥","∟","≡",
+		"⊿","∞","√","?","┌","┬","┐","├",
+		"┼","┤","└","┴","┘","─","│","═",
+		"╞","╪","╡","╱","╲","╳","╭","╮",
+		"╰","╯","▁","▂","▃","▄","▅","▆",
+		"▇","█","▏","▎","▍","▌","▋","▊",
+		"▉","▔","〡","〢","〣","〤","〥","〦",
+		"〧","〨","〩","十","卄","卅","Ⅰ","Ⅱ",
+		"Ⅲ","Ⅳ","Ⅴ","Ⅵ","Ⅶ","Ⅷ","Ⅸ","Ⅹ",
+		"Α","Β","Γ","Δ","Ε","Ζ","Η","Θ",
+		"Ι","Κ","Λ","Μ","Ν","Ξ","Ο","Π",
+		"Ρ","Σ","Τ","Υ","Φ","Χ","Ψ","Ω",
+		"α","β","γ","δ","ε","ζ","η","θ",
+		"ι","κ","λ","μ","ν","ξ","ο","π",
+		"ρ","σ","τ","υ","φ","χ","ψ","ω"};
+	int i, all = 216;
 
-    pci->nTotalChoice = 0;
-    for(i = 0; i< all; i++){
-        memcpy( pci->totalChoiceStr[ pci->nTotalChoice],
-                chibuf[i], sizeof(char)*2) ;
-        pci->totalChoiceStr[ pci->nTotalChoice][2] = '\0' ;
-        pci->nTotalChoice++; 
-    }  
-    pai->avail[1].len = 1;
-    pai->avail[1].id = -1;  
-    pai->nAvail = all;
-    pai->currentAvail = 1;
-    pci->nChoicePerPage = (selectAreaLen - 5) / ( 2 + 3) ;
-    if(pci->nChoicePerPage > MAX_SELKEY) pci->nChoicePerPage = MAX_SELKEY ;
-    pci->nPage = CEIL_DIV(pci->nTotalChoice, pci->nChoicePerPage) ;
-    pci->pageNo = 0 ;
-    pci->isSymbol = 1;
-    return ZUIN_ABSORB;
+	pci->nTotalChoice = 0;
+	for ( i = 0; i< all; i++ ) {
+		ueStrNCpy( pci->totalChoiceStr[ pci->nTotalChoice ],
+				chibuf[i], 1, 1);
+		pci->nTotalChoice++; 
+	}  
+	pai->avail[ 0 ].len = 1;
+	pai->avail[ 0 ].id = -1;  
+	pai->nAvail = 1;
+	pai->currentAvail = 0;
+	pci->nChoicePerPage = (selectAreaLen - 5) / ( 2 + 3) ;
+	if ( pci->nChoicePerPage > MAX_SELKEY ) {
+		pci->nChoicePerPage = MAX_SELKEY;
+	}
+	pci->nPage = CEIL_DIV( pci->nTotalChoice, pci->nChoicePerPage );
+	pci->pageNo = 0;
+	pci->isSymbol = 1;
+	return ZUIN_ABSORB;
 }
 
 static int InternalSpecialSymbol(
@@ -147,7 +148,7 @@ static int InternalSpecialSymbol(
 		int nSpecial, char keybuf[], char *chibuf[] )
 {
 	int i, rtn = ZUIN_IGNORE; /* very strange , and very difficult to understand */
-        int kbtype;
+	int kbtype;
 
 	for ( i = 0; i < nSpecial; i++ ) {
 		if ( key == keybuf[ i ] ) {
@@ -159,18 +160,24 @@ static int InternalSpecialSymbol(
 				( pgdata->chiSymbolBufLen - pgdata->chiSymbolCursor ) );
 
 			pgdata->chiSymbolBuf[ pgdata->chiSymbolCursor ].wch = (wchar_t) 0;
-			pgdata->chiSymbolBuf[ pgdata->chiSymbolCursor ].s[ 0 ] = 
-				chibuf[ i ][ 0 ];
-			pgdata->chiSymbolBuf[ pgdata->chiSymbolCursor ].s[ 1 ] = 
-				chibuf[ i ][ 1 ];
+			ueStrNCpy( (char *) pgdata->chiSymbolBuf[ pgdata->chiSymbolCursor ].s,
+			           chibuf[ i ],
+				   1, 1);
+			/* Save Symbol Key */
+			memmove( &( pgdata->symbolKeyBuf[ pgdata->chiSymbolCursor + 1 ] ),
+				&( pgdata->symbolKeyBuf[ pgdata->chiSymbolCursor ] ),
+				sizeof( pgdata->symbolKeyBuf[0] ) * 
+				( pgdata->chiSymbolBufLen - pgdata->chiSymbolCursor ) );
+			pgdata->symbolKeyBuf[ pgdata->chiSymbolCursor ] = key;
+
 			pgdata->chiSymbolCursor++;
 			pgdata->chiSymbolBufLen++;
 			pgdata->bUserArrCnnct[ pgdata->cursor ] = 0;
 			/* reset Zuin data */
-                        /* Don't forget the kbtype */
-                        kbtype = pgdata->zuinData.kbtype;
+			/* Don't forget the kbtype */
+			kbtype = pgdata->zuinData.kbtype;
 			memset( &( pgdata->zuinData ), 0, sizeof( ZuinData ) );
-                        pgdata->zuinData.kbtype = kbtype;
+			pgdata->zuinData.kbtype = kbtype;
 			break;
 		}
 	}
@@ -187,15 +194,45 @@ int SpecialSymbolInput( int key, ChewingData *pgdata )
 	};
 
 	static char *chibuf[] = {
-		"u", "v", "y", "z", "B", "A", "G", "F", "C",
-		"", "I", "I", "", "C", "H", "s", "", "",
-		"]", "^", "", "", "", "@", "U", "H",
-		"A", "C", "F"
+		"「", "」", "『", "』", "、", "，", "：", "；", "。",
+		"～", "！", "＠", "＃", "＄", "％", "︿", "＆", "＊",
+		"（", "）", "﹍", "＋", "＝", "＼", "｜", "？",
+		"，", "。", "；"
 	};
 	static int nSpecial = 29;
 	return InternalSpecialSymbol( key, pgdata, nSpecial, keybuf, chibuf );
 }
 
+int FullShapeSymbolInput( int key, ChewingData *pgdata )
+{
+	int rtn;
+	static char keybuf[] = {
+		'0','1','2','3','4','5','6','7','8','9',
+		'a','b','c','d','e','f','g','h','i','j',
+		'k','l','m','n','o','p','q','r','s','t',
+		'u','v','w','x','y','z','A','B','C','D',
+		'E','F','G','H','I','J','K','L','M','N',
+		'O','P','Q','R','S','T','U','V','W','X',
+		'Y','Z',' ','\"','\'','/','<','>','`','[',
+		']','{','}','+','-'
+	};
+	static char* chibuf[] = {
+		"０","１","２","３","４","５","６","７","８","９",
+		"ａ","ｂ","ｃ","ｄ","ｅ","ｆ","ｇ","ｈ","ｉ","ｊ",
+		"ｋ","ｌ","ｍ","ｎ","ｏ","ｐ","ｑ","ｒ","ｓ","ｔ",
+		"ｕ","ｖ","ｗ","ｘ","ｙ","ｚ","Ａ","Ｂ","Ｃ","Ｄ",
+		"Ｅ","Ｆ","Ｇ","Ｈ","Ｉ","Ｊ","Ｋ","Ｌ","Ｍ","Ｎ",
+		"Ｏ","Ｐ","Ｑ","Ｒ","Ｓ","Ｔ","Ｕ","Ｖ","Ｗ","Ｘ",
+		"Ｙ","Ｚ","　","”","’","／","＜","＞","‵","〔",
+		"〕","｛","｝","＋","－"
+	};
+	static int nSpecial = sizeof(keybuf) / sizeof(char);
+	rtn = InternalSpecialSymbol( key, pgdata, nSpecial, keybuf, chibuf );
+	if ( rtn == ZUIN_IGNORE )
+		rtn = SpecialSymbolInput( key, pgdata );
+	return (rtn == ZUIN_IGNORE ? SYMBOL_KEY_ERROR : SYMBOL_KEY_OK);
+}
+
 int SpecialEtenSymbolInput( int key, ChewingData *pgdata )
 {
 	static char keybuf[] = {
@@ -205,24 +242,27 @@ int SpecialEtenSymbolInput( int key, ChewingData *pgdata )
 	};
 
 	static char *chibuf[] = {
-		"z","s","{","","q","r", "K","B","C","",
-		"u","q","t","i","j","","", "X", "x", "F","G",
-		"|","r","}", "", "m", "n" ,"w", "A","D","H"
+		"┌","┬","┐","□","〈","〉", "…","、","。","※",
+		"├","┼","┤","【","】","◇","○", "—", "│", "；","：",
+		"└","┴","┘", "ˇ", "《", "》" ,"─", "，","．","？"
 	};
 	static int nSpecial = 31;
 	return InternalSpecialSymbol( key, pgdata, nSpecial, keybuf, chibuf );
 }
 
-int SymbolChoice(ChewingData *pgdata, int sel_i){
+int SymbolChoice( ChewingData *pgdata, int sel_i )
+{
         int kbtype;
-	pgdata->chiSymbolCursor -- ;
-	memmove( &(pgdata->chiSymbolBuf[pgdata->chiSymbolCursor]),
-                 &(pgdata->chiSymbolBuf[pgdata->chiSymbolCursor] ) ,
-                 sizeof(wch_t)*
-                 (pgdata->chiSymbolBufLen - pgdata->chiSymbolCursor) ) ;
-	pgdata->chiSymbolBuf[pgdata->chiSymbolCursor].wch = (wchar_t) 0 ;
-	pgdata->chiSymbolBuf[pgdata->chiSymbolCursor].s[0] = pgdata->choiceInfo.totalChoiceStr[sel_i][0];
-	pgdata->chiSymbolBuf[pgdata->chiSymbolCursor].s[1] = pgdata->choiceInfo.totalChoiceStr[sel_i][1];
+	if ( pgdata->choiceInfo.isSymbol == 1 )
+		pgdata->chiSymbolCursor --;
+	memmove( &(pgdata->chiSymbolBuf[ pgdata->chiSymbolCursor ] ),
+                 &(pgdata->chiSymbolBuf[ pgdata->chiSymbolCursor ] ) ,
+                 sizeof(wch_t) *
+                 (pgdata->chiSymbolBufLen - pgdata->chiSymbolCursor) );
+	pgdata->chiSymbolBuf[pgdata->chiSymbolCursor].wch = (wchar_t) 0;
+	ueStrNCpy( (char *) pgdata->chiSymbolBuf[pgdata->chiSymbolCursor].s,
+	           pgdata->choiceInfo.totalChoiceStr[sel_i],
+		   1, 1);
 	pgdata->chiSymbolCursor ++ ; 
 	pgdata->bUserArrCnnct[pgdata->cursor] = 0;
 	ChoiceEndChoice(pgdata);
@@ -236,14 +276,15 @@ int SymbolChoice(ChewingData *pgdata, int sel_i){
 int SemiSymbolInput(int key, ChewingData *pgdata)
 {
     static char keybuf[] = {' ', '1'} ;
-    static char *chibuf[] = {"@", ""} ;
+    static char *chibuf[] = {"　", "符"} ;
     static int nSpecial = 2 ;
     return InternalSpecialSymbol( key, pgdata, nSpecial, keybuf, chibuf );
 }
 
 int SymbolInput( int key, ChewingData *pgdata )
 {
-	if ( isprint( (char) key ) ) { /* other character was ignored */
+	if ( isprint( (char) key ) && /* other character was ignored */
+	     (pgdata->chiSymbolBufLen < MAX_PHONE_SEQ_LEN) ) { /* protect the buffer */
 		memmove(
 			&( pgdata->chiSymbolBuf[ pgdata->chiSymbolCursor + 1 ] ),
 			&( pgdata->chiSymbolBuf[ pgdata->chiSymbolCursor ] ),
@@ -251,6 +292,14 @@ int SymbolInput( int key, ChewingData *pgdata )
 
 		pgdata->chiSymbolBuf[ pgdata->chiSymbolCursor ].wch = (wchar_t) 0;
 		pgdata->chiSymbolBuf[ pgdata->chiSymbolCursor ].s[ 0 ] = (char) key;
+
+		/* Save Symbol Key */
+		memmove( &( pgdata->symbolKeyBuf[ pgdata->chiSymbolCursor + 1 ] ),
+			&( pgdata->symbolKeyBuf[ pgdata->chiSymbolCursor ] ),
+			sizeof( pgdata->symbolKeyBuf[0] ) * 
+			( pgdata->chiSymbolBufLen - pgdata->chiSymbolCursor ) );
+			pgdata->symbolKeyBuf[ pgdata->chiSymbolCursor ] = toupper( key );
+	
 		pgdata->chiSymbolCursor++;
 		pgdata->chiSymbolBufLen++;
 		pgdata->bUserArrCnnct[ pgdata->cursor ] = 0;
@@ -267,7 +316,7 @@ static int CompInterval( const IntervalType *a, const IntervalType *b )
 	return ( a->to - b->to );
 }
 
-int FindIntervalFrom( int from, IntervalType inte[], int nInte )
+static int FindIntervalFrom( int from, IntervalType inte[], int nInte )
 {
 	int i;
 
@@ -287,17 +336,19 @@ int WriteChiSymbolToBuf( wch_t csBuf[], int csBufLen, ChewingData *pgdata )
 			 * Workaround to avoid different initialization behavior 
 			 * among Win32 and Unix-like OSs.
 			 */
-			memset( &( csBuf[ i ].s ), 0, WCH_SIZE );
-			memcpy( csBuf[ i ].s, &( pgdata->phrOut.chiBuf[ phoneseq_i ] ), 2 );
-			phoneseq_i += 2;
+			memset( &( csBuf[ i ].s ), 0, MAX_UTF8_SIZE + 1 );
+			ueStrNCpy( (char *) csBuf[ i ].s,
+			           &( pgdata->phrOut.chiBuf[ phoneseq_i ] ), 
+				   1, 1);
+			phoneseq_i += ueBytesFromChar( pgdata->phrOut.chiBuf[ phoneseq_i ] );
 		}
 		else 
-			csBuf[ i ].wch = pgdata->chiSymbolBuf[ i ].wch;
+			csBuf[ i ] = pgdata->chiSymbolBuf[ i ];
 	}
 	return 0;
 }
 
-int CountReleaseNum( ChewingData *pgdata )
+static int CountReleaseNum( ChewingData *pgdata )
 {
 	int remain, i;
 
@@ -323,7 +374,7 @@ int CountReleaseNum( ChewingData *pgdata )
 	return 1;
 }
 
-void KillFromLeft( ChewingData *pgdata, int nKill )
+static void KillFromLeft( ChewingData *pgdata, int nKill )
 {
 	int i;
 
@@ -349,13 +400,17 @@ void CleanAllBuf( ChewingData *pgdata )
 	pgdata->chiSymbolCursor = 0;
 	/* 7 */
 	memset( pgdata->bUserArrCnnct, 0, sizeof( pgdata->bUserArrCnnct ) );
+
+	pgdata->phrOut.nNumCut = 0;
+
+	memset( pgdata->symbolKeyBuf, 0, sizeof( pgdata->symbolKeyBuf ) );
 }
 
 int ReleaseChiSymbolBuf( ChewingData *pgdata, ChewingOutput *pgo )
 {
 	int throwEnd;
 	uint16 bufPhoneSeq[ MAX_PHONE_SEQ_LEN + 1 ];
-	char bufWordSeq[ MAX_PHONE_SEQ_LEN * 2 + 1 ];
+	char bufWordSeq[ MAX_PHONE_SEQ_LEN * 3 + 1 ];
 
 	throwEnd = CountReleaseNum( pgdata );
 
@@ -370,8 +425,7 @@ int ReleaseChiSymbolBuf( ChewingData *pgdata, ChewingOutput *pgo )
 		/* Add to userphrase */
 		memcpy( bufPhoneSeq, pgdata->phoneSeq, sizeof( uint16 ) * throwEnd );
 		bufPhoneSeq[ throwEnd ] = (uint16) 0;
-		memcpy( bufWordSeq, pgdata->phrOut.chiBuf, sizeof( char ) * throwEnd * 2 );
-		bufWordSeq[ throwEnd * 2 ] = '\0';
+		ueStrNCpy( bufWordSeq, pgdata->phrOut.chiBuf, throwEnd, 1 );
 		UserUpdatePhrase( bufPhoneSeq, bufWordSeq );
 
 		KillFromLeft( pgdata, throwEnd );
@@ -382,7 +436,7 @@ int ReleaseChiSymbolBuf( ChewingData *pgdata, ChewingOutput *pgo )
 void AutoLearnPhrase( ChewingData *pgdata )
 {
 	uint16 bufPhoneSeq[ MAX_PHONE_SEQ_LEN + 1 ];
-	char bufWordSeq[ MAX_PHONE_SEQ_LEN * 2 + 1 ];
+	char bufWordSeq[ MAX_PHONE_SEQ_LEN * 3 + 1 ];
 	int i, from, len;
 
 	for ( i = 0; i < pgdata->nPrefer; i++ ) {
@@ -390,11 +444,9 @@ void AutoLearnPhrase( ChewingData *pgdata )
 		len = pgdata->preferInterval[i].to - from;
 		memcpy( bufPhoneSeq, &pgdata->phoneSeq[ from ], sizeof( uint16 ) * len );
 		bufPhoneSeq[ len ] = (uint16) 0;
-		memcpy( 
-			bufWordSeq, 
-			&pgdata->phrOut.chiBuf[ from * 2 ], 
-			sizeof( char ) * len * 2 );
-		bufWordSeq[ len * 2 ] = '\0';
+		ueStrNCpy( bufWordSeq,
+		           ueStrSeek( (char *) &pgdata->phrOut.chiBuf, from ),
+		           len, 1);
 		UserUpdatePhrase( bufPhoneSeq, bufWordSeq );
 	}
 }
@@ -444,7 +496,7 @@ int AddChi( uint16 phone, ChewingData *pgdata )
 }
 
 #ifdef ENABLE_DEBUG
-void ShowChewingData( ChewingData *pgdata )
+static void ShowChewingData( ChewingData *pgdata )
 {
 	int i ;
 
@@ -489,7 +541,6 @@ void ShowChewingData( ChewingData *pgdata )
 		pgdata->bSelect, 
 		pgdata->bCaseChange );
 }
-
 #endif
 
 int CallPhrasing( ChewingData *pgdata )
@@ -618,7 +669,7 @@ static void ShiftInterval( ChewingOutput *pgo, ChewingData *pgdata )
 	}
 }
 
-int MakeOutput( ChewingOutput *pgo, ChewingData *pgdata )
+static int MakeOutput( ChewingOutput *pgo, ChewingData *pgdata )
 {
 	int chi_i, chiSymbol_i, i ;
 
@@ -633,16 +684,14 @@ int MakeOutput( ChewingOutput *pgo, ChewingData *pgdata )
 		if ( pgdata->chiSymbolBuf[ chiSymbol_i ].wch == (wchar_t) 0 ) { 
 			/* is Chinese, then copy from the PhrasingOutput "phrOut" */
 			pgo->chiSymbolBuf[ chiSymbol_i ].wch = (wchar_t) 0;
-			memcpy(
-				pgo->chiSymbolBuf[ chiSymbol_i ].s,
-				&( pgdata->phrOut.chiBuf[ chi_i ] ),
-				2 );
-			chi_i += 2;
+			ueStrNCpy( (char *) pgo->chiSymbolBuf[ chiSymbol_i ].s,
+			           &( pgdata->phrOut.chiBuf[ chi_i ] ),
+			           1, 1 );
+			chi_i += ueBytesFromChar( pgo->chiSymbolBuf[ chiSymbol_i ].s[0] );
 		}
 		else {
 			/* is Symbol */
-			pgo->chiSymbolBuf[ chiSymbol_i ].wch = 
-				pgdata->chiSymbolBuf[ chiSymbol_i ].wch;
+			pgo->chiSymbolBuf[ chiSymbol_i ] = pgdata->chiSymbolBuf[ chiSymbol_i ];
 		}
 	}
 
@@ -655,28 +704,35 @@ int MakeOutput( ChewingOutput *pgo, ChewingData *pgdata )
 	pgo->chiSymbolCursor = pgdata->chiSymbolCursor;
 	
 	/*  fill zuinBuf */
-        if(pgdata->zuinData.kbtype >= KB_HANYU_PINYING) {
-		char *p = pgdata->zuinData.pinYingData.keySeq;
+        if ( pgdata->zuinData.kbtype >= KB_HANYU_PINYIN ) {
+		char *p = pgdata->zuinData.pinYinData.keySeq;
+		/* 
+		 * Copy from old content in zuinBuf
+		 * NOTE: No Unicode transformation here.
+		 */
 		for ( i = 0; i< ZUIN_SIZE; i++) {
 			int j;
-			for(j = 0; j < 2; j++) {
-				if(p[0]) {
-					pgo->zuinBuf[i].s[j] = p[0];
+			for ( j = 0; j < 2; j++ ) {
+				if ( p[0] ) {
+					pgo->zuinBuf[ i ].s[ j ] = p[ 0 ];
 					p++;
-				} else {
-					pgo->zuinBuf[i].s[j] = '\0';
+				} 
+				else {
+					pgo->zuinBuf[ i ].s[ j ] = '\0';
 				}
 			}
-			pgo->zuinBuf[i].s[2]='\0';
+			pgo->zuinBuf[ i ].s[ 2 ] = '\0';
 		}
 	} else {
 		for ( i = 0; i < ZUIN_SIZE; i++ ) { 
 			if ( pgdata->zuinData.pho_inx[ i ] != 0 ) {
-				memcpy( 
-					pgo->zuinBuf[ i ].s, 
-					& ph_pho[ i ][ pgdata->zuinData.pho_inx[ i ] * 2 ], 
-					2 ); 
-				pgo->zuinBuf[ i ].s[ 2 ] = '\0';
+				/* Here we should use (zhuin_tab[i] + 2) to
+				 * skip the 2 space characters at 
+				 * zhuin_tab[0] and zhuin_tab[1]. */
+				ueStrNCpy( (char *) pgo->zuinBuf[ i ].s,
+				           ueStrSeek( (char *) (zhuin_tab[ i ] + 2),
+						      pgdata->zuinData.pho_inx[ i ] - 1 ),
+				           1, 1);
 			}
 			else
 				pgo->zuinBuf[ i ].wch = (wchar_t) 0;
@@ -717,10 +773,9 @@ int AddSelect( ChewingData *pgdata, int sel_i )
 	nSelect = pgdata->nSelect;
 
 	/* change "selectStr" , "selectInterval" , and "nSelect" of ChewingData */
-	memcpy(
-		pgdata->selectStr[ nSelect ],
-		pgdata->choiceInfo.totalChoiceStr[ sel_i ],
-		length * 2 ) ;
+	ueStrNCpy( pgdata->selectStr[ nSelect ],
+			pgdata->choiceInfo.totalChoiceStr[ sel_i ],
+			length, 1 );
 	pgdata->selectInterval[ nSelect ].from = pgdata->cursor;
 	pgdata->selectInterval[ nSelect ].to = pgdata->cursor + length;
 	pgdata->nSelect++;
@@ -755,7 +810,7 @@ void RemoveSelectElement( int i, ChewingData *pgdata )
 	strcpy( pgdata->selectStr[ i ], pgdata->selectStr[ pgdata->nSelect ] );
 }
 
-int ChewingKillSelectIntervalAcross( int cursor, ChewingData *pgdata )
+static int ChewingKillSelectIntervalAcross( int cursor, ChewingData *pgdata )
 {
 	int i;
 	for ( i = 0; i < pgdata->nSelect; i++ ) {
@@ -768,7 +823,7 @@ int ChewingKillSelectIntervalAcross( int cursor, ChewingData *pgdata )
 	return 0;
 }
 
-int KillCharInSelectIntervalAndBrkpt( ChewingData *pgdata, int cursorToKill )
+static int KillCharInSelectIntervalAndBrkpt( ChewingData *pgdata, int cursorToKill )
 {
 	int i; 
 
@@ -810,6 +865,7 @@ int ChewingKillChar(
 		pgdata->nPhoneSeq--;
 		pgdata->cursor -= minus;
 	}
+	pgdata->symbolKeyBuf[ chiSymbolCursorToKill ] = 0;
 	memmove( 
 		& pgdata->chiSymbolBuf[ chiSymbolCursorToKill ],
 		& pgdata->chiSymbolBuf[ chiSymbolCursorToKill + 1 ], 
@@ -832,3 +888,112 @@ int IsPreferIntervalConnted( int cursor, ChewingData *pgdata )
 	return 0;
 }
 
+int OpenSymbolChoice( ChewingData *pgdata )
+{
+	static char *symbol_buf[][50] = {
+		{ "0","ø", 0 },
+		{ "[", "「", "『", "《", "〈", "【", "〔", 0 },
+		{ "]", "」", "』", "》", "〉", "】", "〕", 0 },
+		{ "{", "｛", 0 },
+		{ "}", "｝", 0 },
+		{ "<", "，", "←", 0 },
+		{ ">", "。", "→", "．", 0 },
+		{ "?", "？","¿", 0 },
+		{ "!", "！", "①", "➀", "Ⅰ","¡", 0 },
+		{ "@", "＠", "②", "➁", "Ⅱ", "⊕", "⊙", "㊣", "﹫", 0 },
+		{ "#", "＃", "③", "➂", "Ⅲ", "﹟", 0 },
+		{ "$", "＄", "④", "➃", "Ⅳ", "€", "﹩", "￠", "∮","￡", "￥", 0 },
+		{ "%", "％", "⑤", "➄", "Ⅴ", 0 },
+		{ "^", "︿", "⑥", "➅", "Ⅵ", "﹀", "︽", "︾", 0 },
+		{ "&", "＆", "⑦", "➆", "Ⅶ", "﹠", 0 },
+		{ "*", "＊", "⑧", "➇", "Ⅷ", "×", "※", "╳", "﹡", "☯","☆", "★", 0 },
+		{ "(", "（", "⑨", "➈", "Ⅸ", 0 },
+		{ ")", "）", "⑩", "➉", "Ⅹ", 0 },
+		{ "_", "＿", "…", "‥", "←", "→", "﹍", "﹉", "ˍ", "￣", "–", "—", "¯", "﹊", "﹎", "﹏", "﹣", "－", 0 },
+		{ "+", "＋", "±", "﹢", "✙", "✚", "✛", "✜", "✝", "✞", "✟", 0 },
+		{ "=", "＝", "≒", "≠", "≡", "≦", "≧", "﹦", 0},
+		{ "`", "』", "『", "′", "‵", 0 },
+		{ "~", "～", 0 },
+		{ ":", "：", "；", "︰", "﹕", 0 },
+		{ "\"", "；", 0 },
+		{ "\'", "、", "…", "‥", 0 },
+		{ "\\", "＼", "↖", "↘", "﹨", 0 },
+		{ "-", "－", "＿", "￣", "¯", "ˍ", "–", "—", "‥", "…",
+			"←","→","╴","﹉","﹊","﹍","﹎","﹏","﹣", 0 },
+		{ "/","／","÷","↗","↙","∕", 0 },
+		{ "|","↑", "↓", "∣", "∥", "︱", "︳", "︴" ,0 },
+		{ "A", "Ⓐ", "Å","Α", "α", "├", "╠", "╟", "╞", 0 },
+		{ "B", "Ⓑ", "Β", "β","∵", 0 },
+		{ "C", "Ⓒ", "Χ", "χ", "┘", "╯", "╝", "╜", "╛","㏄","℃","㎝","♣","♧","©" ,0 },
+		{ "D", "Ⓓ", "Δ", "δ", "◇", "◆", "┤", "╣", "╢", "╡","♦", 0 },
+		{ "E", "Ⓔ", "Ε", "ε", "┐", "╮", "╗", "╓", "╕", 0 },
+		{ "F", "Ⓕ", "Φ", "ψ", "│", "║", "℉","♀", 0 },
+		{ "G", "Ⓖ", "Γ", "γ", 0 },
+		{ "H", "Ⓗ", "Η", "η","♥","♡", 0 },
+		{ "I", "Ⓘ", "Ι", "ι", 0 },
+		{ "J", "Ⓙ", "φ", 0 },
+		{ "K", "Ⓚ", "Κ", "κ","㎞", "㏎", 0 },
+		{ "L", "Ⓛ", "Λ", "λ","㏒", "㏑", 0 },
+		{ "M", "Ⓜ", "Μ", "μ","♂","ℓ","㎎", "㏕", "㎜","㎡", 0 },
+		{ "N", "Ⓝ", "Ν", "ν","№", 0 },
+		{ "O", "Ⓞ", "Ο", "ο", 0 },
+		{ "P", "Ⓟ", "Π", "π", 0 },
+		{ "Q", "Ⓠ", "Θ", "θ","Д","┌", "╭", "╔", "╓", "╒","۞", 0 },
+		{ "R", "Ⓡ", "Ρ", "ρ", "─", "═" ,"®" , 0 },
+		{ "S", "Ⓢ", "Σ", "σ","∴","□","■","┼", "╬", "╪", "╫","∫","§","♠","♤", 0 },
+		{ "T", "Ⓣ", "Τ", "τ","θ","△","▲","▽","▼","™","⊿", "™", 0 },
+		{ "U", "Ⓤ", "Υ", "υ","μ","∪", "∩", 0 },
+		{ "V", "Ⓥ", 0 },
+		{ "W", "Ⓦ", "Ω", "ω", "┬", "╦", "╤", "╥", 0 },
+		{ "X", "Ⓧ", "Ξ", "ξ", "┴", "╩", "╧", "╨", 0 },
+		{ "Y", "Ⓨ", "Ψ", 0 },
+		{ "Z", "Ⓩ", "Ζ", "ζ", "└", "╰", "╚", "╙", "╘", 0 },
+	};
+	int i, symbol_buf_len = 56;
+	char **pBuf;
+	ChoiceInfo *pci = &( pgdata->choiceInfo );
+	pci->oldCursor = pgdata->cursor;
+	pci->oldChiSymbolCursor = pgdata->chiSymbolCursor;
+
+	/* see if there is some word in the cursor position */
+	if ( pgdata->chiSymbolCursor == pgdata->chiSymbolBufLen )
+		pgdata->chiSymbolCursor--;
+	if ( pgdata->symbolKeyBuf[ pgdata->chiSymbolCursor ] == '1' ) {
+		pgdata->bSelect = 1;
+		HaninSymbolInput( pci, &( pgdata->availInfo ), 
+				pgdata->phoneSeq, pgdata->config.selectAreaLen );
+		pci->isSymbol = 2;
+		return 0;
+	}
+	for ( i = 0; i < symbol_buf_len; i++ ) {
+		if ( symbol_buf[i][0][0] == pgdata->symbolKeyBuf[ pgdata->chiSymbolCursor ] ) {
+			pBuf = symbol_buf[i];
+			break;
+		}
+	}
+	if ( i == symbol_buf_len ) {
+		ChoiceEndChoice( pgdata );
+		return 0;
+	}
+	pci->nTotalChoice = 0;
+	for ( i = 1; pBuf[i]; i++ ) {
+		ueStrNCpy( pci->totalChoiceStr[pci->nTotalChoice], 
+				pBuf[i], ueStrLen( pBuf[i] ), 1 );
+		pci->nTotalChoice++; 
+	}
+
+	pci->nChoicePerPage = (pgdata->config.selectAreaLen - 5) / ( 2 + 3) ;
+	if ( pci->nChoicePerPage > MAX_SELKEY )
+		pci->nChoicePerPage = MAX_SELKEY ;
+	pci->nPage = CEIL_DIV( pci->nTotalChoice, pci->nChoicePerPage );
+	pci->pageNo = 0;
+	pci->isSymbol = 2;
+
+	pgdata->bSelect = 1;
+	pgdata->availInfo.nAvail = 1;
+	pgdata->availInfo.currentAvail = 0;
+	pgdata->availInfo.avail[ 0 ].id = -1;
+	pgdata->availInfo.avail[ 0 ].len = 1;     
+	return 0;
+}
+
